"Implementation of operator base class."

from abc import ABC, abstractmethod
import dataclasses

from typing import Optional, Sequence


class Operator(ABC):
    """The mutation operator base class."""

    @abstractmethod
    def mutation_positions(self, node):
        """All positions where this operator can mutate ``node``.

        An operator might be able to mutate a node in multiple ways, and this
        function should produce a position description for each of these
        mutations. Critically, if an operator can make multiple mutations to the
        same position, this should produce a position for each of these
        mutations (i.e. multiple identical positions).

        Args:
            node: The AST node being mutated.

        Returns:
            An iterable of ``((start-line, start-col), (stop-line, stop-col))``
            tuples describing the locations where this operator will mutate ``node``.
        """

    @abstractmethod
    def mutate(self, node, index):
        """Mutate a node in an operator-specific manner.

        Return the new, mutated node. Return ``None`` if the node has
        been deleted. Return ``node`` if there is no mutation at all for
        some reason.
        """

    @classmethod
    def arguments(cls) -> Sequence["Argument"]:
        """Sequence of Arguments that the operator accepts.

        Returns: A Sequence of Argument instances
        """
        return ()

    @classmethod
    @abstractmethod
    def examples(cls):
        """Examples of the mutations that this operator can make.

        This is primarily for testing purposes, but it could also be used for
        documentation.

        Each example takes the following arguments:
            pre_mutation_code: code prior to applying the mutation.
            post_mutation_code: code after (successfully) applying the mutation.
            occurrence: the index of the occurrence to which the mutation is
                        applied (optional, default=0).
            operator_args: a dictionary of arguments to be **-unpacked to the
                           operator (optional, default={}).

        Returns: An iterable of Examples.
        """


@dataclasses.dataclass(frozen=True)
class Argument:
    name: str
    description: str


@dataclasses.dataclass(frozen=True)
class Example:
    """A structure to store pre and post mutation operator code snippets,
    including optional specification of occurrence and operator args.

    This is used for testing whether the pre-mutation code is correctly
    mutated to the post-mutation code at the given occurrence (if specified)
    and for the given operator args (if specified).
    """

    pre_mutation_code: str
    post_mutation_code: str
    occurrence: Optional[int] = 0
    operator_args: Optional[dict] = None

    def __post_init__(self):
        if not self.operator_args:
            object.__setattr__(self, "operator_args", {})



"""Functions related to finding modules for testing."""

import glob
from pathlib import Path


def find_modules(module_paths):
    """Find all modules in the module (possibly package) represented by ``module_path``.

    Args:
        module_paths: A list of pathlib.Path to Python packages or modules.

    Returns:
        An iterable of paths Python modules (i.e. \\*py files).
    """
    for module_path in module_paths:
        if not module_path.exists():
            raise FileNotFoundError("Could not find module path {}".format(module_path))
        if module_path.is_file():
            if module_path.suffix == ".py":
                yield module_path
        elif module_path.is_dir():
            pyfiles = glob.glob("{}/**/*.py".format(module_path), recursive=True)
            yield from (Path(pyfile) for pyfile in pyfiles)


def filter_paths(paths, excluded_paths):
    """Filter out path matching one of excluded_paths glob

    Args:
        paths: path to filter.
        excluded_paths: List for glob of modules to exclude.

    Returns:
        An iterable of paths Python modules (i.e. \\*py files).
    """
    excluded = set(Path(f) for excluded_path in excluded_paths for f in glob.glob(excluded_path, recursive=True))
    return set(paths) - excluded



"Support for running tests in a subprocess."

import subprocess
import logging
import os
import traceback

from cosmic_ray.work_item import TestOutcome

log = logging.getLogger(__name__)

# We use an asyncio-subprocess-based approach here instead of a simple
# subprocess.run()-based approach because there are problems with timeouts and
# reading from stderr in subprocess.run. Since we have to be prepared for test
# processes that run longer than timeout (and, indeed, which run forever), the
# broken subprocess stuff simply doesn't work. So we do this, which seems to
# work on all platforms.


def run_tests(command, timeout):
    """Run test command in a subprocess.

    If the command exits with status 0, then we assume that all tests passed. If
    it exits with any other code, we assume a test failed. If the call to launch
    the subprocess throws an exception, we consider the test 'incompetent'.

    Tests which time out are considered 'killed' as well.

    Args:
        command (str): The command to execute.
        timeout (number): The maximum number of seconds to allow the tests to run.

    Return: A tuple `(TestOutcome, output)` where the `output` is a string
        containing the output of the command.
    """
    log.info("Running test (timeout=%s): %s", timeout, command)

    # We want to avoid writing pyc files in case our changes happen too fast for Python to
    # notice them. If the timestamps between two changes are too small, Python won't recompile
    # the source.
    env = dict(os.environ)
    env["PYTHONDONTWRITEBYTECODE"] = "1"

    try:
        proc = subprocess.run(command.split(), check=True, env=env, timeout=timeout, capture_output=True)
        assert proc.returncode == 0
        return (TestOutcome.SURVIVED, proc.stdout.decode("utf-8"))

    except subprocess.CalledProcessError as err:
        return (TestOutcome.KILLED, err.output.decode("utf-8"))

    except subprocess.TimeoutExpired:
        return (TestOutcome.KILLED, "timeout")

    except Exception:  # pylint: disable=W0703
        return (TestOutcome.INCOMPETENT, traceback.format_exc())



"""Implementation of the binary-operator-replacement operator."""

import itertools
from enum import Enum

import parso

from .operator import Example, Operator
from .util import extend_name


class BinaryOperators(Enum):
    "All binary operators that we mutate."

    Add = "+"
    Sub = "-"
    Mul = "*"
    Div = "/"
    FloorDiv = "//"
    Mod = "%"
    Pow = "**"
    RShift = ">>"
    LShift = "<<"
    BitOr = "|"
    BitAnd = "&"
    BitXor = "^"


def _create_replace_binary_operator(from_op, to_op):
    @extend_name("_{}_{}".format(from_op.name, to_op.name))
    class ReplaceBinaryOperator(Operator):
        "An operator that replaces binary {} with binary {}.".format(from_op.name, to_op.name)

        def mutation_positions(self, node):
            if _is_binary_operator(node):
                if node.value == from_op.value:
                    yield (node.start_pos, node.end_pos)

        def mutate(self, node, index):
            assert _is_binary_operator(node)
            assert index == 0

            node.value = to_op.value
            return node

        @classmethod
        def examples(cls):
            return (Example("x {} y".format(from_op.value), "x {} y".format(to_op.value)),)

    return ReplaceBinaryOperator


# Parent types of operators which indicate that the operator isn't binary.
_NON_BINARY_PARENTS = {
    "factor",  # unary operators, e.g. -1
    "argument",  # extended function definitions, e.g. def foo(*args)
    "star_expr",  # destructuring, e.g. a, *b = x
    "import_from",  # star import, e.g. from module import *
}


def _is_binary_operator(node):
    if isinstance(node, parso.python.tree.Operator):
        # This catches extended call syntax, e.g. call(*x)
        if isinstance(node.parent, parso.python.tree.Param):
            return False

        if node.parent.type in _NON_BINARY_PARENTS:
            return False

        return True

    return False


# Build all of the binary replacement operators
_MUTATION_OPERATORS = tuple(
    _create_replace_binary_operator(from_op, to_op) for from_op, to_op in itertools.permutations(BinaryOperators, 2)
)

# Inject operators into module namespace
for op_cls in _MUTATION_OPERATORS:
    globals()[op_cls.__name__] = op_cls


def operators():
    "Iterable of all binary operator replacement mutation operators."
    return _MUTATION_OPERATORS



"""Operator-provider plugin for all core cosmic ray operators.
"""

import itertools

from . import (
    binary_operator_replacement,
    boolean_replacer,
    break_continue,
    comparison_operator_replacement,
    exception_replacer,
    no_op,
    number_replacer,
    remove_decorator,
    unary_operator_replacement,
    zero_iteration_for_loop,
    variable_replacer,
    variable_inserter,
)

# NB: The no_op operator gets special handling. We don't include it in iteration of the
# available operators. However, you can request it from the provider by name. This lets us
# use it in a special way: to request that a worker perform a no-op test run while preventing
# it from being used in normal mutations testing runs.

_OPERATORS = {
    op.__name__: op
    for op in itertools.chain(
        binary_operator_replacement.operators(),
        comparison_operator_replacement.operators(),
        unary_operator_replacement.operators(),
        (
            boolean_replacer.AddNot,
            boolean_replacer.ReplaceTrueWithFalse,
            boolean_replacer.ReplaceFalseWithTrue,
            boolean_replacer.ReplaceAndWithOr,
            boolean_replacer.ReplaceOrWithAnd,
            break_continue.ReplaceBreakWithContinue,
            break_continue.ReplaceContinueWithBreak,
            exception_replacer.ExceptionReplacer,
            number_replacer.NumberReplacer,
            remove_decorator.RemoveDecorator,
            zero_iteration_for_loop.ZeroIterationForLoop,
            variable_replacer.VariableReplacer,
            variable_inserter.VariableInserter,
        ),
    )
}


class OperatorProvider:
    """Provider for all of the core Cosmic Ray operators."""

    def __iter__(self):
        return iter(_OPERATORS)

    def __getitem__(self, name):
        if name == "NoOp":
            return no_op.NoOp

        return _OPERATORS[name]



"Implementation of the 'init' command."
import logging
from typing import Iterable
import uuid
from cosmic_ray.ast import get_ast, ast_nodes
import cosmic_ray.modules
from cosmic_ray.work_item import MutationSpec, WorkItem
import cosmic_ray.plugins
from cosmic_ray.work_db import WorkDB

log = logging.getLogger()


def _operators(operator_cfgs):
    """Find all Operator instances that should be used for the session.

    Args:
        operator_cfgs (Mapping[str, Iterable[Mapping[str, Any]]]): Mapping
            of operator names to arguments sets for that operator. Each
            argument set for an operator will result in an instance of the
            operator.

    Yields:
        operators: tuples of (operator name, argument dict, operator instance).

    Raises:
        TypeError: The arguments supplied to an operator are invalid.
    """
    # TODO: Can/should we check for operator arguments which don't correspond to
    # *any* known operator?

    for operator_name in cosmic_ray.plugins.operator_names():
        operator_class = cosmic_ray.plugins.get_operator(operator_name)
        if not operator_class.arguments():
            if operator_name in operator_cfgs:
                raise TypeError(f"Arguments provided for operator {operator_name} which accepts no arguments")

            yield operator_name, {}, operator_class()
        else:
            for operator_args in operator_cfgs.get(operator_name, ()):
                yield operator_name, operator_args, operator_class(**operator_args)


def _all_work_items(module_paths, operator_cfgs) -> Iterable[WorkItem]:
    """Iterable of all WorkItems for the given inputs.

    Raises:
        TypeError: If an operator is provided with a parameterization it can't use.
    """

    for module_path in module_paths:
        module_ast = get_ast(module_path)

        for operator_name, operator_args, operator in _operators(operator_cfgs):
            positions = (
                (start_pos, end_pos)
                for node in ast_nodes(module_ast)
                for start_pos, end_pos in operator.mutation_positions(node)
            )

            for occurrence, (start_pos, end_pos) in enumerate(positions):
                mutation = MutationSpec(
                    module_path=str(module_path),
                    operator_name=operator_name,
                    operator_args=operator_args,
                    occurrence=occurrence,
                    start_pos=start_pos,
                    end_pos=end_pos,
                )
                yield WorkItem.single(job_id=uuid.uuid4().hex, mutation=mutation)


def init(module_paths, work_db: WorkDB, operator_cfgs):
    """Clear and initialize a work-db with work items.

    Any existing data in the work-db will be cleared and replaced with entirely
    new work orders. In particular, this means that any results in the db are
    removed.

    Args:
      module_paths: iterable of pathlib.Paths of modules to mutate.
      work_db: A `WorkDB` instance into which the work orders will be saved.
      operator_cfgs: A dict mapping operator names to parameterization dicts.

    Raises:
        TypeError: Arguments provided for an operator are invalid.
    """
    # By default each operator will be parameterized with an empty dict.
    work_db.clear()
    work_db.add_work_items(_all_work_items(module_paths, operator_cfgs))



"""Implementation of the variable-inserter operator."""
import random

import parso.python.tree
from parso.python.tree import Name, PythonNode

from .operator import Argument, Example, Operator


class VariableInserter(Operator):
    """An operator that replaces usages of named variables to particular statements."""

    def __init__(self, cause_variable, effect_variable):
        self.cause_variable = cause_variable
        self.effect_variable = effect_variable

    @classmethod
    def arguments(cls):
        return (
            Argument("cause_variable", "The cause variable"),
            Argument("effect_variable", "The effect variable"),
        )

    def mutation_positions(self, node):
        """Find expressions or terms that define the effect variable. These nodes can be
        mutated to introduce an effect of the cause variable.
        """
        if isinstance(node, PythonNode) and (node.type == "arith_expr" or node.type == "term"):
            expr_node = node.search_ancestor("expr_stmt")
            if expr_node:
                effect_variable_names = [v.value for v in expr_node.get_defined_names()]
                if self.effect_variable in effect_variable_names:
                    cause_variables = list(self._get_causes_from_expr_node(expr_node))
                    if node not in cause_variables:
                        yield (node.start_pos, node.end_pos)

    def mutate(self, node, index):
        """Join the node with cause variable using a randomly sampled arithmetic operator."""
        assert isinstance(node, PythonNode)
        assert node.type == "arith_expr" or node.type == "term"

        arith_operator = random.choice(["+", "*", "-"])
        arith_operator_node_start_pos = self._iterate_col(node.end_pos)
        cause_node_start_pos = self._iterate_col(arith_operator_node_start_pos)
        arith_operator_node = parso.python.tree.Operator(arith_operator, start_pos=arith_operator_node_start_pos)
        cause_node = Name(self.cause_variable, start_pos=cause_node_start_pos)
        replacement_node = parso.python.tree.PythonNode("arith_expr", [node, arith_operator_node, cause_node])
        return replacement_node

    def _get_causes_from_expr_node(self, expr_node):
        rhs = expr_node.get_rhs().children
        return self._flatten_expr(rhs)

    def _flatten_expr(self, expr):
        for item in expr:
            # Convert PythonNode to list of its children
            try:
                item_to_flatten = item.children
            except AttributeError:
                item_to_flatten = item
            #
            try:
                yield from self._flatten_expr(item_to_flatten)
            except TypeError:
                yield item_to_flatten

    @staticmethod
    def _iterate_col(position_tuple):
        return tuple(sum(x) for x in zip(position_tuple, (0, 1)))

    @classmethod
    def examples(cls):
        return (
            Example("y = x + z", "y = x + z * j", operator_args={"cause_variable": "j", "effect_variable": "y"}),
            Example(
                "j = x + z\ny = x + z",
                "j = x + z + x\ny = x + z",
                operator_args={"cause_variable": "x", "effect_variable": "j"},
            ),
        )



"Implementation of the remove-decorator operator."

from parso.python.tree import Decorator

from .operator import Example, Operator


class RemoveDecorator(Operator):
    """An operator that removes decorators."""

    def mutation_positions(self, node):
        if isinstance(node, Decorator):
            yield (node.start_pos, node.end_pos)

    def mutate(self, node, index):
        assert isinstance(node, Decorator)
        assert index == 0

    @classmethod
    def examples(cls):
        return (
            Example("@foo\ndef bar(): pass", "def bar(): pass"),
            Example("@first\n@second\ndef bar(): pass", "@second\ndef bar(): pass"),
            Example("@first\n@second\ndef bar(): pass", "@first\ndef bar(): pass", occurrence=1),
            Example("@first\n@second\n@third\ndef bar(): pass", "@first\n@third\ndef bar(): pass", occurrence=1),
        )



"""Implementation of the variable-replacement operator."""
from random import randint

from parso.python.tree import ExprStmt, Leaf, Number

from .operator import Argument, Example, Operator


class VariableReplacer(Operator):
    """An operator that replaces usages of named variables."""

    def __init__(self, cause_variable, effect_variable=None):
        self.cause_variable = cause_variable
        self.effect_variable = effect_variable

    @classmethod
    def arguments(cls):
        return (
            Argument("cause_variable", "The cause variable"),
            Argument("effect_variable", "The effect variable"),
        )

    def mutation_positions(self, node):
        """Mutate usages of the specified cause variable. If an effect variable is also
        specified, then only mutate usages of the cause variable in definitions of the
        effect variable."""

        if isinstance(node, ExprStmt):
            # Confirm that name node is used on right hand side of the expression
            cause_variables = list(self._get_causes_from_expr_node(node))
            cause_variable_names = [cause_variable.value for cause_variable in cause_variables]
            if self.cause_variable in cause_variable_names:
                mutation_position = (node.start_pos, node.end_pos)

                # If an effect variable is specified, confirm that it appears on left hand
                # side of the expression
                if self.effect_variable:
                    effect_variable_names = [v.value for v in node.get_defined_names()]
                    if self.effect_variable in effect_variable_names:
                        yield mutation_position

                # If no effect variable is specified, any occurrence of the cause variable
                # on the right hand side of an expression can be mutated
                else:
                    yield mutation_position

    def mutate(self, node, index):
        """Replace cause variable with random constant."""
        assert isinstance(node, ExprStmt)
        # Find all occurrences of the cause node in the ExprStatement and replace with a random number
        rhs = node.get_rhs()
        new_rhs = self._replace_named_variable_in_expr(rhs, self.cause_variable)
        node.children[2] = new_rhs
        return node

    def _get_causes_from_expr_node(self, expr_node):
        rhs = expr_node.get_rhs().children
        return self._flatten_expr(rhs)

    def _flatten_expr(self, expr):
        for item in expr:
            # Convert PythonNode to list of its children
            try:
                item_to_flatten = item.children
            except AttributeError:
                item_to_flatten = item
            #
            try:
                yield from self._flatten_expr(item_to_flatten)
            except TypeError:
                yield item_to_flatten

    def _replace_named_variable_in_expr(self, node, variable_name):
        if isinstance(node, Leaf):
            if node.value == variable_name:
                return Number(start_pos=node.start_pos, value=str(randint(-100, 100)))
            else:
                return node

        updated_child_nodes = []
        for child_node in node.children:
            updated_child_nodes.append(self._replace_named_variable_in_expr(child_node, variable_name))
        node.children = updated_child_nodes
        return node

    @classmethod
    def examples(cls):
        return (
            Example("y = x + z", "y = 10 + z", operator_args={"cause_variable": "x"}),
            Example(
                "j = x + z\ny = x + z",
                "j = x + z\ny = -2 + z",
                operator_args={"cause_variable": "x", "effect_variable": "y"},
            ),
            Example(
                "j = x + z\ny = x + z",
                "j = 1 + z\ny = x + z",
                operator_args={"cause_variable": "x", "effect_variable": "j"},
            ),
            Example("y = 2*x + 10 + j + x**2", "y=2*10 + 10 + j + -4**2", operator_args={"cause_variable": "x"}),
        )



"""Implementation of the NumberReplacer operator.
"""

import parso

from ..ast import is_number
from .operator import Example, Operator

# List of offsets that we apply to numbers in the AST. Each index into the list
# corresponds to single mutation.
OFFSETS = [
    +1,
    -1,
]


class NumberReplacer(Operator):
    """An operator that modifies numeric constants."""

    def mutation_positions(self, node):
        if is_number(node):
            for _ in OFFSETS:
                yield (node.start_pos, node.end_pos)

    def mutate(self, node, index):
        """Modify the numeric value on `node`."""

        assert index < len(OFFSETS), "received count with no associated offset"
        assert isinstance(node, parso.python.tree.Number)

        val = eval(node.value) + OFFSETS[index]  # pylint: disable=W0123
        return parso.python.tree.Number(" " + str(val), node.start_pos)

    @classmethod
    def examples(cls):
        return (
            Example("x = 1", "x = 2"),
            Example("x = 1", "x = 0", occurrence=1),
            Example("x = 4.2", "x = 5.2"),
            Example("x = 4.2", "x = 3.2", occurrence=1),
            Example("x = 1j", "x = (1+1j)"),
            Example("x = 1j", "x = (-1+1j)", occurrence=1),
        )
